#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sstream>

#include "opensslutil.h"
#include "dcisec.h"

namespace AuthN {

using namespace AuthN::OpenSSL;

  static std::string tostring(unsigned int val) {
    std::stringstream ss;
    ss << val;
    return ss.str();
  }

  CredentialsRequest::CredentialsRequest(void) :
      context_(NULL), csrctx_(NULL), keyctx_(NULL), valid_from_(-1), valid_till_(-1) {
  }

  CredentialsRequest::CredentialsRequest(const CredentialsRequest& arg) :
      context_(NULL), csrctx_(NULL), keyctx_(NULL), valid_from_(-1), valid_till_(-1) {
    if(arg.context_) context_ = &arg.context_->Copy();
    Assign(arg);
  }

  CredentialsRequest& CredentialsRequest::operator=(const CredentialsRequest& arg) {
    Assign(arg);
    return *this;
  }

  Status CredentialsRequest::Assign(const CredentialsRequest& arg) {
    if(!csrctx_) return (last_error_ = Status(-1, "Failed to find CertContext object"));
    if(!keyctx_) return (last_error_ = Status(-1, "Failed to find KeyContext object"));
    std::string req, key;
    csrctx_->csrToPEM(req);
    keyctx_->privateToPEM(key);

    return AssignRequest(req, key);
  }

  CredentialsRequest::CredentialsRequest(const AuthN::Context& ctx) : csrctx_(NULL), keyctx_(NULL) {
    context_ = &ctx.Copy();
    csrctx_ = new AuthN::OpenSSL::CSRContext(*context_);
    keyctx_ = new AuthN::OpenSSL::KeyContext(*context_);
  }

  CredentialsRequest::~CredentialsRequest() {
    if(context_) delete context_;
    if(csrctx_) delete csrctx_; 
    if(keyctx_) delete keyctx_;
  }

  CredentialsRequest& CredentialsRequest::Copy(void) const {
    CredentialsRequest* req = new CredentialsRequest(context_?*context_:Context());
    std::string req_str, priv_str;
    GetRequest(req_str);
    GetPrivateKey(priv_str);
    req->AssignRequest(req_str, priv_str);
    return *req;    
  }

  Status CredentialsRequest::GetStatus(void) const {
    return last_error_;
  }

  AuthN::Status CredentialsRequest::MakeKeys(int bits) {
    if(!keyctx_) return (last_error_ = Status(-1, "Failed to find KeyContext object"));
    return (keyctx_->createKey(bits));
  }

  Status CredentialsRequest::GetPrivateKey(std::string& str, bool encrypt) const{
    if(!keyctx_) return Status(-1, "KeyContext is empty");
    keyctx_->privateToPEM(str, encrypt);
    return Status(0); //TODO: return more status;

/*
    if(!priv_key_) {
      return Status(-1,"There is no private key assigned");
    }

    BIO* priv_bio = BIO_new(BIO_s_mem());
    if(!priv_bio) {
      return Status(-1,"Failed to create BIO for private key: "+GetOpenSSLError());
    }
    if(!encrypt) {
      if(!PEM_write_bio_PrivateKey(priv_bio, priv_key_, NULL,
            NULL, 0, NULL, NULL)) {
        BIO_free_all(priv_bio);
        return Status(-1,"Failed to write private key: "+GetOpenSSLError());
      }
    }
    else {
      const EVP_CIPHER *cipher = EVP_des_ede3_cbc();
      if(!PEM_write_bio_PrivateKey(priv_bio, priv_key_, cipher,
            NULL, 0, &ContextOpenSSLUI::passwordcb, context_)) {
        BIO_free_all(priv_bio);
        return Status(-1,"Failed to write encrypted private key: "+GetOpenSSLError());
      }
    }

    for(;;) {
      char s[256];
      int l = BIO_read(priv_bio,s,sizeof(s));
      if(l <= 0) break;
      str.append(s,l);
    }
    BIO_free_all(priv_bio);
    return Status(0);
*/
  }

  Status CredentialsRequest::GetPrivateKey(std::ostream& o, bool encrypt) const{
    std::string s;
    Status err = GetPrivateKey(s,encrypt);
    o << s;
    return err;
  }

  Status CredentialsRequest::GetPublicKey(std::string& str) const{
    if(!keyctx_) return Status(-1, "KeyContext is empty");
    keyctx_->publicToPEM(str);
    return Status(0); //TODO: return more status;

/*
    if(!pub_key_) {
      return Status(-1,"There is no public key assigned");
    }

    BIO* pub_bio = BIO_new(BIO_s_mem());
    if(!pub_bio) {
      return Status(-1,"Failed to create BIO for public key: "+GetOpenSSLError());
    }
    if(!PEM_write_bio_PUBKEY(pub_bio, pub_key_)) {
      BIO_free_all(pub_bio);
      return Status(-1,"Failed to write public key: "+GetOpenSSLError());
    }
    for(;;) {
      char s[256];
      int l = BIO_read(pub_bio,s,sizeof(s));
      if(l <= 0) break;
      str.append(s,l);
    }
    BIO_free_all(pub_bio);
    return Status(0);
*/
  }

  Status CredentialsRequest::GetPublicKey(std::ostream& o) const{
    std::string s;
    Status err = GetPublicKey(s);
    o << s;
    return err;
  }

  AuthN::Status CredentialsRequest::MakeRequest(void) {
    if(!csrctx_) return (last_error_ = Status(-1, "Failed to find CSRContext object"));
    if(!keyctx_) return (last_error_ = Status(-1, "Failed to find KeyContext object"));

    AuthN::OpenSSL::CertificateOptions req_opts;
    req_opts.setSubject(subject_name_);
    req_opts.setAsUser();
    AuthN::Utils::Time start, end;
    if(valid_from_ != -1) start = valid_from_;
    else start = AuthN::Utils::Time();
    if(valid_till_ != -1) end = valid_till_;
    if(start < end)req_opts.setValidityPeriod(start, end);

    if(!(csrctx_->createRequest(req_opts, *keyctx_))) 
      return (last_error_ = Status(-1, "Failed to create X509 request"));
    return Status(0);

/*
    req_ = X509_REQ_new();
    if(!req_) { return Status(-1); }
    if(!make_request(*context_, req_, pub_key_, priv_key_, false, conf_file, subject_name_)) { return Status(-1); }
    return Status(0);
*/
//    if(!X509_REQ_set_version(req_,3L)) { return Status(-1); } 
//    X509_NAME *name = NULL;
//    unsigned long chtype = MBSTRING_ASC;  //TODO
//    name = parse_name((char*)(dn.c_str()), chtype, 0);
//    X509_REQ_set_subject_name(req, name);
//    X509_NAME_free(name);

//    if(!X509_REQ_set_pubkey(req_, pub_key_)) { return Status(-1); }
//    const EVP_MD *digest = (EVP_MD*)EVP_sha1();
//    if(!X509_REQ_sign(req_, priv_key_, digest)) { return Status(-1); }
//    return Status(0);
  }

  //const X509_REQ* CredentialsRequest::GetRequest(void) const {
  //  return req_;
  //}

  Status CredentialsRequest::GetRequest(std::string& str) const {
    if(!csrctx_) return Status(-1, "Failed to find CSRContext object"); 
    csrctx_->csrToPEM(str);
    return Status(0);

/*
    if(!req_) {
      return Status(-1,"There is no request assigned");
    }

    BIO* req_bio = BIO_new(BIO_s_mem());
    if(!req_bio) {
      return Status(-1,"Failed to create BIO for request: "+GetOpenSSLError());
    }

    if(!PEM_write_bio_X509_REQ(req_bio, req_)){
      BIO_free_all(req_bio);
      return Status(-1,"Failed to write request: "+GetOpenSSLError());
    }

    for(;;) {
      char s[256];
      int l = BIO_read(req_bio,s,sizeof(s));
      if(l <= 0) break;
      str.append(s,l);
    }

    BIO_free_all(req_bio);
    return Status(0);
*/
  }

  Status CredentialsRequest::GetRequest(std::ostream& o) const {
    std::string s;
    Status err = GetRequest(s);
    o << s;
    return err;
  }

  AuthN::Status CredentialsRequest::AssignKeys(std::istream& priv) {
    std::string priv_str;

    std::getline<char>(priv, priv_str, 0);
    if (!priv) return Status(-1);
    return(AssignKeys(priv_str));
  }

  AuthN::Status CredentialsRequest::AssignKeys(const std::string& priv) {
    if(!keyctx_) return (last_error_ = Status(-1, "Failed to find KeyContext object"));
    if(!(keyctx_->privateFromStr(priv))) return (last_error_ = Status(-1, "Failed to assign private key"));
    return Status(0);

/*
    const unsigned char* pub_str = (const unsigned char*)(pub.c_str());
    pub_key_ = d2i_PublicKey(EVP_PKEY_RSA, NULL, &pub_str, pub.length());

    BIO* priv_bio;
    priv_bio = BIO_new_mem_buf((unsigned char*)(priv.c_str()), priv.length());
    unsigned char *passphrase = NULL; //TODO: input passphrase
    priv_key_ = PEM_read_bio_PrivateKey(priv_bio, NULL, 0, passphrase);
    BIO_free(priv_bio);

    return Status(0);
*/
  }

  AuthN::Status CredentialsRequest::AssignRequest(std::istream& req, std::istream& priv) {
    std::string req_str;
    std::string priv_str;

    std::getline<char>(req, req_str, 0);
    if (!req) return Status(-1);

    std::getline<char>(priv, priv_str, 0);
    if (!priv) return Status(-1);
    return(AssignRequest(req_str, priv_str));
  }

  AuthN::Status CredentialsRequest::AssignRequest(const std::string& req, const std::string& priv) {
    if(!csrctx_) return (last_error_ = Status(-1, "Failed to find CertContext object"));
    if(!keyctx_) return (last_error_ = Status(-1, "Failed to find KeyContext object"));
    if(!(csrctx_->csrFromPEM(req))) return (last_error_ = Status(-1, "Failed to load CSR"));
    if(!priv.empty() && !(keyctx_->privateFromStr(priv))) return (last_error_ = Status(-1, "Failed to load private key"));
    subject_name_ = csrctx_->getProps()->subject;
    valid_from_ = csrctx_->getProps()->start.GetTime();
    valid_till_ = csrctx_->getProps()->end.GetTime();
    return Status(0);
  }

  AuthN::Status CredentialsRequest::SetSubjectName(const std::string sn) {
    subject_name_ = sn;
    //TODO: parse the sn, and compose it into RFC2253
    return Status(0);
  }

  std::string CredentialsRequest::GetSubjectName(void) const {
    return subject_name_;
  }

  time_t CredentialsRequest::GetValidFrom(void) const {
    return valid_from_;
  }

  void CredentialsRequest::SetValidFrom(time_t sec) {
    valid_from_ = sec;
  }

  time_t CredentialsRequest::GetValidTill(void) const {
    return valid_till_;
  }

  void CredentialsRequest::SetValidTill(time_t sec) {
    valid_till_ = sec;
  }

  Status CredentialsRequest::GetExtension(int pos, Credentials::Extension& ext) const {
    if(!csrctx_) return false;
    return csrctx_->getCSRExtension(pos, ext);

/*
    if(!req_) {
      return Status(-1,"There is no request assigned");
    }

    std::string oid;
    bool crit = false;
    std::string ext_val;

    STACK_OF(X509_EXTENSION)* exts = NULL;
    exts = X509_REQ_get_extensions(req_);
    if(!exts) {
      return Status(-1,"There are no extensions: "+GetOpenSSLError());
    }
    int num = sk_X509_EXTENSION_num(exts);
    if((num > 0) && (pos < num) && (pos >= 0)) {
      X509_EXTENSION* x509_ext = sk_X509_EXTENSION_value(exts, pos);
      if(!x509_ext) {
        return Status(-1,"Failed to obtain requested extension: "+GetOpenSSLError());
      }
      if(!parse_extension(x509_ext, crit, oid, ext_val)) {
        return Status(-1,"Failed to parse requested extension: "+GetOpenSSLError());
      }
    }
    else if(pos >=num) {
      return Status(-1,"There is no extension at requested position");
    }

    ext.critical = crit;
    ext.oid = oid;
    ext.value = ext_val;
    return Status(0);
*/
  }

  Status CredentialsRequest::GetExtension(const std::string& name, Credentials::Extension& ext) const {
    if(!csrctx_) return false;
    return csrctx_->getCSRExtension(name, ext);

/*
    if(!req_) {
      return Status(-1,"There is no request assigned");
    }

    std::string oid;
    bool crit = false;
    std::string ext_val;

    STACK_OF(X509_EXTENSION)* exts = NULL;
    exts = X509_REQ_get_extensions(req_);
    if(!exts) {
      return Status(-1,"There are no extensions: "+GetOpenSSLError());
    }
    int num = sk_X509_EXTENSION_num(exts);
    for(int i = 0; i < num; i++) {
      X509_EXTENSION* x509_ext = sk_X509_EXTENSION_value(exts, i);
      if(!x509_ext) continue; // or fail?
      const char *extname = OBJ_nid2sn(OBJ_obj2nid(X509_EXTENSION_get_object(x509_ext)));
      if (name == extname) {
        if(!parse_extension(x509_ext, crit, oid, ext_val)) {
          return Status(-1,"Failed to parse requested extension: "+GetOpenSSLError());
        }
        ext.critical = crit;
        ext.oid = oid;
        ext.value = ext_val;
        return Status(0);
      }
    }
    return Status(-1,"There is no extension with requested name");
*/
  }

  Status CredentialsRequest::AddExtension(Credentials::Extension& ext) {
    if(!csrctx_) return false;
    return csrctx_->setCSRExtension(ext);

/*
    if(!req_) {
      return (last_error_ = Status(-1,"There is no request assigned"));
    }

    STACK_OF(X509_EXTENSION)* exts = X509_REQ_get_extensions(req_);
    if(!exts) {
      return (last_error_ = Status(-1,"Failed to obtain extensions: "+GetOpenSSLError()));
    }

    if(ext.oid.empty()) {
      return (last_error_ = Status(-1,"New extension is missing OID"));
    }

    // TODO: check for result of OBJ_txt2nid
    X509_EXTENSION* x509_ext = X509V3_EXT_conf_nid(NULL, NULL, OBJ_txt2nid((char*)(ext.oid.c_str())), (char*)(ext.value.c_str()));
    if(!x509_ext) {
      return (last_error_ = Status(-1,"Failed to create extensions: "+GetOpenSSLError()));
    }
    if(!sk_X509_EXTENSION_push(exts, x509_ext)) {
      X509_EXTENSION_free(x509_ext);
      return (last_error_ = Status(-1,"Failed to add extensions: "+GetOpenSSLError()));
    }
    return Status(0);
*/
  }

  Status CredentialsRequest::GetAttributes(const std::string& name, std::list<std::string>& attrs) const {
    return Status(-1,"Not implemented");
  }

  Status CredentialsRequest::AddAttributes(const std::string& name, std::list<std::string>& attrs) {
    return Status(-1,"Not implemented");
  }

  CACredentialsRequest::CACredentialsRequest(const AuthN::Context& ctx) : CredentialsRequest(ctx), certctx_(NULL) {
  }

  CACredentialsRequest::~CACredentialsRequest(void) {
    if(certctx_) delete certctx_;
  }

  CredentialsRequest& CACredentialsRequest::Copy(void) const {
    CACredentialsRequest* req = new CACredentialsRequest(context_?*context_:Context());
    std::string req_str, priv_key_str;
    GetRequest(req_str);
    GetPrivateKey(priv_key_str);
    req->CredentialsRequest::AssignRequest(req_str, priv_key_str);
    std::string cert;
    if(certctx_) { certctx_->certToPEM(cert); req->Assign(cert); }
    return *req;
  }

  AuthN::Status CACredentialsRequest::MakeRequest(void) {
    if(!csrctx_) {
      context_->Log(Context::LogError,"Failed to find CSRContext object");
      return (last_error_ = Status(-1, "Failed to find CSRContext object"));
    }
    if(!keyctx_) {
      context_->Log(Context::LogError,"Failed to find KeyContext object");
      return (last_error_ = Status(-1, "Failed to find KeyContext object"));
    }

    AuthN::OpenSSL::CertificateOptions req_opts;
    req_opts.setSubject(subject_name_);
    req_opts.setAsCA();
    AuthN::Utils::Time start, end;
    if(valid_from_ != -1) start = valid_from_;
    else start = AuthN::Utils::Time();
    if(valid_till_ != -1) end = valid_till_;
    if(start < end)req_opts.setValidityPeriod(start, end);

    if(!(csrctx_->createRequest(req_opts, *keyctx_))) {
      context_->Log(Context::LogError,"Failed to create X509 request");
      return (last_error_ = Status(-1, "Failed to create X509 request"));
    }
    return Status(0);

  }

  Status CACredentialsRequest::GetCertificate(std::string& str) {
    if(!certctx_) return (last_error_ = Status(-1, "The CertContext if empty"));
    certctx_->certToPEM(str);
    if(str.empty()) return (last_error_ = Status(-1,"The X509 certificate is empty"));
    return Status(0);

/*
    if(!cert_) {
      return Status(-1,"There is no certificate assigned");
    }
    BIO *out = BIO_new(BIO_s_mem());
    if(!out) {
      return Status(-1,"Failed to create BIO for certificate: "+GetOpenSSLError());
    }
    if(!PEM_write_bio_X509(out,cert_)) {
      BIO_free_all(out);
      return Status(-1,"Failed to write certificate: "+GetOpenSSLError());
    }

    for(;;) {
      char s[256];
      int l = BIO_read(out,s,sizeof(s));
      if(l <= 0) break;
      str.append(s,l);
    }
    BIO_free_all(out);
    return Status(0);
*/
  }

  Status CACredentialsRequest::GetCertificate(std::ostream& o) {
    std::string s;
    Status err = GetCertificate(s);
    o << s;
    return err;
  }

  AuthN::Status CACredentialsRequest::Assign(const std::string& cert) {
    if(!certctx_) return (last_error_ = Status(-1, "Failed to find CertContext object"));
    if(!(certctx_->certFromPEM(cert))) return (last_error_ = Status(-1, "Failed to load X509 Certificate"));
    return Status(0);     
  }

  ProxyCredentialsRequest::ProxyCredentialsRequest(const AuthN::Context& ctx) : CredentialsRequest(ctx) { };

  ProxyCredentialsRequest::~ProxyCredentialsRequest(void) { };

  CredentialsRequest& ProxyCredentialsRequest::Copy(void) const {
    ProxyCredentialsRequest* req = new ProxyCredentialsRequest(context_?*context_:Context());
    std::string req_str, priv_key_str;
    GetRequest(req_str);
    GetPrivateKey(priv_key_str);
    req->CredentialsRequest::AssignRequest(req_str, priv_key_str);
    return *req;

/*
    ProxyCredentialsRequest* req = new ProxyCredentialsRequest(context_?*context_:Context());
    req->CredentialsRequest::Assign(req_, pub_key_, priv_key_);
    return *req;
*/
  }

  bool ProxyCredentialsRequest::GetLimited(void) {
    return false;
  }

  void ProxyCredentialsRequest::SetLimited(bool) {
  }

  Status ProxyCredentialsRequest::GetPolicy(Credentials::Extension& policy) {
    const CertContextProps* props = csrctx_->getProps();
    policy.oid = props->proxyPolicy.oid;
    policy.critical = props->proxyPolicy.critical;
    policy.value = props->proxyPolicy.value;
    return policy.oid.empty()?false:true;

  }

  Status ProxyCredentialsRequest::SetPolicy(const Credentials::Extension& policy) {
    // The method could cover two cases: 
    // 1. On the request side, SetPolicy should be called before MakeRequest, 
    //    so that the policy will be insert into X509_REQ
    // 2. On the signer side, the call of SetPolicy will only cause the change 
    //    of CertContextProps.
    CertContextProps* props = &(csrctx_->props_);
    props->proxyPolicy.value = policy.value;
     
    return Status(0);
  }

}//AuthN


